@@ -33,7 +33,7 @@ import vcon.security
 import vcon.filter_plugins
 import vcon.accessors
 
-__version__ = "0.3"
+__version__ = "0.3.1"
 
 def build_logger(name : str) -> logging.Logger:
   logger = logging.getLogger(name)
@@ -683,11 +683,9 @@ class Vcon():
         party_index = self.set_party_parameter("mailto", email_address[1])
         self.set_party_parameter("name", email_address[0], party_index)
         parties_found = [party_index]
-        if(sender_index is None):
-          sender_index = parties_found[0]
 
-      elif(sender_index is None):
-          sender_index = party_index
+      if(sender_index is None):
+        sender_index = parties_found[0]
 
       party_indices.extend(parties_found)
 
